import cv2
from skimage.measure import profile_line # conda install scikit-image, conda install chardet
from skimage.measure import find_contours
from skimage import filters
from skimage.draw import disk 
import tkinter as tk
from tkinter import ttk
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from PIL import Image, ImageTk
from scipy.stats import norm
from scipy.signal import find_peaks,peak_widths
import os
import random

from matplotlib import cm
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk\

table_path = 'D:/Users/James/Desktop/Catherine Science 2023/Python_Apps/GalaxyClassifier/Main/gz2_imgtable.csv'

source_path = 'C:/GZ2/images_gz2'
output_path = 'C:/GZ2/images_cntr'

class contourApp(tk.Tk):
    def __init__(self,master=None):
        super().__init__(master)

        self.raw_images = []
        self.class_ids = []
        self.asset_ids = []

        # UI Setup
        menubar = tk.Menu(master=self)
        fileMenu = tk.Menu(menubar,tearoff=0)
        menubar.add_cascade(label="File", menu=fileMenu)
        fileMenu.add_command(label='Preprocess and Save All', command=self.saveProcessedImages)
        fileMenu.add_separator()
        fileMenu.add_command(label='Exit', command=self.destroy)
        self.config(menu=menubar)

        self.wm_title('Image Preprocessing: Contour')
        self.wm_geometry('848x500')

        mainFrame=tk.Frame(master=self)
        mainFrame.pack(fill=tk.BOTH,side=tk.TOP,expand=True)

        displayFrame=tk.Frame(master=mainFrame)
        displayFrame.pack(fill=tk.BOTH,side=tk.TOP,expand=True)

        controlFrame=tk.Frame(master=mainFrame)
        controlFrame.pack(fill=tk.BOTH,side=tk.TOP,expand=False)

        self.sourceCanvas = tk.Canvas(master=displayFrame,width=424,height=424,bg="black")
        self.sourceCanvas.pack(side=tk.LEFT,expand=False)

        self.procCanvas = tk.Canvas(master=displayFrame,width=424,height=424)
        self.procCanvas.pack(side=tk.LEFT,expand=False)
        #self.procCanvas.create_line(0,0,423,423,fill='red',width=2)

        #procFigure = plt.Figure()
        #self.procPlot = procFigure.add_subplot()
        #self.procCanvas = FigureCanvasTkAgg(procFigure,displayFrame)
        #self.procCanvas.get_tk_widget().pack(fill=tk.BOTH,side=tk.LEFT,expand=False)

        self.progBar = ttk.Progressbar(controlFrame,orient='horizontal',mode='indeterminate')
        self.progBar.pack(fill=tk.BOTH,side=tk.LEFT,expand=True)

        self.backButton = tk.Button(controlFrame,text='Back',command=self.backImage)
        self.backButton.pack(fill=tk.BOTH,side=tk.LEFT,expand=True)
        #self.backButton['state'] = 'disabled'

        self.forwardButton = tk.Button(controlFrame,text='Forward',command=self.forwardImage)
        self.forwardButton.pack(fill=tk.BOTH,side=tk.LEFT,expand=True)
        #self.forwardButton['state'] = 'disabled'

        startLabel = tk.Label(controlFrame,text="Start:")
        startLabel.pack(fill=tk.BOTH,side=tk.LEFT,expand=False)
        self.startEntry = tk.Entry(controlFrame)
        self.startEntry.pack(fill=tk.BOTH,side=tk.LEFT,expand=False)
        self.startEntry.insert(0,'0')
        endLabel = tk.Label(controlFrame,text="End:")
        endLabel.pack(fill=tk.BOTH,side=tk.LEFT,expand=False)
        self.endEntry = tk.Entry(controlFrame)
        self.endEntry.pack(fill=tk.BOTH,side=tk.LEFT,expand=False)
        self.endEntry.insert(0,'10')

        processButton=tk.Button(controlFrame,text='Load Source Images',command=self.loadSourceImages)
        processButton.pack(fill=tk.BOTH,side=tk.LEFT,expand=False)

    def loadSourceImages(self):
        # Initialization.
        self.raw_images = []
        self.class_ids = []
        self.asset_ids = []
        self.img_index = 0

        # Load class and asset information.
        gz2_table = pd.read_csv(table_path,encoding='utf-8')
        asset_ids = gz2_table.iloc[:,0].values
        class_ids = gz2_table.iloc[:,2].values

        # Load Pre-processed files.
        #print("Loading source images...")
        #source_list = os.listdir(source_path)

        start = int(self.startEntry.get())
        end = int(self.endEntry.get())
        total = end-start

        #count = 0
        #for asset_id,class_id in zip(asset_ids[start:],class_ids[start:]):
        #    for source in source_list:
        #        if source == str(asset_id)+'.jpg':
        #            #print("Found %s as %s class:%s"%(asset_id,source,class_id))
        #            i_path = source_path+'/'+source
        #            try:
        #                self.raw_images.append(cv2.imread(i_path,cv2.IMREAD_GRAYSCALE))
        #                self.class_ids.append(class_id)
        #                self.asset_ids.append(asset_id)
        #                print("%s"%i_path)
        #                count += 1
        #            except:
        #                print("Skipping missing file-> %s"%(i_path))
        #    if count == total:
        #        break

        i = start
        count = 0
        for asset_id,class_id in zip(asset_ids[start:end],class_ids[start:end]):
        #for asset_id,class_id in zip(asset_ids[start:],class_ids[start:]):
            if (class_id[0] != "A") and ("(" not in class_id):
                i_path = source_path+'/'+str(asset_id)+'.jpg'
                try:
                    img = cv2.imread(i_path,cv2.IMREAD_GRAYSCALE)
                    self.raw_images.append(img)
                    self.class_ids.append(class_id)
                    self.asset_ids.append(asset_id)
                    #print("%i\t%s\t%s"%(i,class_id,i_path))
                    count += 1
                except:
                    print("%i Skipping missing file-> %s"%(i,i_path))
                i += 1
                if i == end:
                    break
            else:
                print("%i\tSkipping:\t%s\t%s"%(i, class_id,i_path))

        print("Loading complete\ti=%i\tcount=%i"%(i,count))

        if(self.raw_images):
            self.showImage(self.raw_images[self.img_index],class_id[self.img_index])

    def saveProcessedImages(self):
        print("Preprocess and Save Started...")
        #files_list = os.listdir(source_path)
        count = 0
        for image,class_id,asset_id in zip(self.raw_images,self.class_ids,self.asset_ids):
            if(image is not None):
                try:
                    max_i,contours = self.get_contours(image)
                except:
                    print("get_contours() issue: asset_id=%i"%(asset_id))
                    print(image)
                    continue
                try:
                    filter = self.circular_filter(contours[max_i],image)
                except:
                    print("circular_filter() issue: asset_id=%i"%(asset_id))
                    break
                if(class_id[0] == "E"):
                    simp_id = 'E'
                elif(class_id[:2] == "Se"):
                    simp_id = 'Se'
                elif(class_id[0] == "S"):
                    simp_id = 'S'
                else:
                    simp_id = 'X'
                f = output_path+'/'+str(asset_id)+"_"+simp_id+'.jpg'
                try:
                    cv2.imwrite(f,filter)
                    count += 1
                    print("Count=%i"%count)
                except:
                    print("Couldn't write: %s"%f)
            else:
                print("Missing image: asset_id=%i"%(asset_id))

        print("Preprocess and Save Complete.")

    def showImage(self, image, class_id):
        # Display the image.
        self.tk_img = ImageTk.PhotoImage(image=Image.fromarray(image)) # MUST USE "self"
        self.sourceCanvas.create_image(0,0,anchor=tk.NW,image=self.tk_img)
        self.sourceCanvas.create_text(5,5,anchor=tk.NW, text='%s'%(class_id),fill='red',font='Helvetica 12 bold',justify='left') 

        self.procCanvas.delete("all")

        # Edge Analysis.
        #edge_image = self.get_edges(image)
        #imax = np.max(edge_image)
        #self.tk_img2 = ImageTk.PhotoImage(image=Image.fromarray(255.0/imax*edge_image))
        #self.procCanvas.create_image(0,0,anchor=tk.NW,image=self.tk_img2)

        # Contour Analysis.
        #self.procCanvas.create_image(0,0,anchor=tk.NW,image=self.tk_img)
        max_i,contours = self.get_contours(image)

        # Other-Object Filtering.
        filter = self.circular_filter(contours[max_i],image)
        self.tk_img2 = ImageTk.PhotoImage(image=Image.fromarray(filter))
        self.procCanvas.create_image(0,0,anchor=tk.NW,image=self.tk_img2)
        self.draw_contours(max_i, contours)


    def get_edges(self, image):
        edge = filters.roberts(image)
        return edge

    def get_contours(self, image):
        contours = np.array(find_contours(image,60),dtype=object)
        max_contour = 0
        max_i = 0
        i = 0
        for contour in contours:
            if len(contour) > max_contour:
                max_i = i;
                max_contour = len(contour)
            i += 1
        return max_i,contours

    def draw_contours(self, max_i, contours):
        i = 0
        for contour in contours:
            x_coords = []
            y_coords = []
            for coords in contour:
                y_coords.append(coords[0])
                x_coords.append(coords[1])
            x0 = x_coords[0]
            y0 = y_coords[0]
            if i != max_i:
                fill = 'red'
            else:
                fill = 'cyan'
            for x,y in zip(x_coords[1:],y_coords[1:]):
                self.procCanvas.create_line(x0,y0,x,y,fill=fill,width=2)
                x0 = x
                y0 = y
            self.procCanvas.create_line(x0,y0,x_coords[-1],y_coords[-1],fill='light green',width=2)
            i += 1

    def circular_filter(self, contour, image):
        # Get image dimensions.
        height,width = image.shape

        # Get filter radius.
        radius = 0
        for coords in contour:
            x = coords[0]-width/2
            y = coords[1]-height/2
            h = np.hypot(x,y)
            if h > radius:
                radius = h
        if (radius > height/2) or (radius > width/2):
            radius = height/2-1
        #radius += 0.25 # add margin
        #print("Filter radius:%f"%radius)

        # Make the filter.
        filter = np.zeros((height,width), dtype=np.uint8)
        rr, cc = disk((height/2,width/2), radius=radius, shape=None)
        filter[rr,cc] = 1.0

        # Apply the filter.
        image_out = filter*image
        #print(image_out.shape)

        return image_out

    def rotate_image(self, image):
        # Normalize image.
        np_image = np.array(image) #/255.0

        # Image properties.
        dims = np_image.shape
        rows = dims[0]
        cols = dims[1]
        #print("Image rows=%d cols=%d"%(rows,cols))

        self.scan_angle = []
        self.scan_peaks = []
        self.scan_widths = []
        for a in range(-90,90):
            # Extract slice profile.
            xi,yi,xf,yf = self.angular_scan(a,rows,cols)
            slice_profile = profile_line(np_image,(yi,xi),(yf,xf))
            # Find profile characteristics.
            peaks, _ = find_peaks(slice_profile)
            results_half = peak_widths(slice_profile, peaks, rel_height=0.5)
            self.max_ind = np.argmax(results_half[0])
            #print("Peaks...")
            #print(peaks)
            #print("Widths...")
            #print(results_half[0])
            self.scan_angle.append(a)
            self.scan_peaks.append(slice_profile[peaks[self.max_ind]])
            self.scan_widths.append(results_half[0][self.max_ind])
            #print("angle=%i argmax=%i width=%f peak=%f"%(a,self.max_ind,results_half[0][self.max_ind],slice_profile[peaks[self.max_ind]]))
        
        # Rotate galaxy image.
        max_width_ind = np.argmax(self.scan_widths)
        print("Rotate-> argmax=%i angle-at-max-width=%i max_width=%f "%(max_width_ind,self.scan_angle[max_width_ind],self.scan_widths[max_width_ind]))
        center = (cols/2,rows/2)
        rotate_matrix = cv2.getRotationMatrix2D(center=center, angle=-self.scan_angle[max_width_ind], scale=1)
        rotated_image = cv2.warpAffine(src=image, M=rotate_matrix, dsize=(cols, rows))

        # Calculate filtering ellipse?

        return rotated_image

    def angular_scan(self,angle,rows,cols):
        # Cross-section line as function of scan angle.
        #angle = 60                  # Angular scan from 90 to -90.
        m = np.tan(angle*np.pi/180) # Slope from angle.
        #print("Scan line: angle=%f slope=%f"%(angle,m))
        if angle == 0:
            xi = cols
            xf = 0
            yf = yi = rows/2
        elif angle > 45:
            y0 = rows/2         # max y.
            x0 = y0/m           # corresponding x.
            xi = int(cols/2+x0) # image offset x-initial.
            xf = int(cols/2-x0) # image offset x-final.
            yi = 0
            yf = cols
        elif angle < -45:
            y0 = rows/2
            x0 = y0/m
            xi = int(cols/2-x0)
            xf = int(cols/2+x0)
            yi = cols
            yf = 0 
        else: #  -45 <= angle <= 45
            xi = cols
            xf = 0
            b0 = m*(-rows/2)
            yi = int(rows/2 + b0)
            yf = int(rows/2 - b0)

        return xi,yi,xf,yf

    def add_noise(img):

        # Getting the dimensions of the image
        row , col = img.shape
        
        # Randomly pick some pixels in the
        # image for coloring them white
        # Pick a random number between 300 and 10000
        number_of_pixels = random.randint(300, 10000)
        for i in range(number_of_pixels):
            
            # Pick a random y coordinate
            y_coord=random.randint(0, row - 1)
            
            # Pick a random x coordinate
            x_coord=random.randint(0, col - 1)
            
            # Color that pixel to white
            img[y_coord][x_coord] = 255
            
        # Randomly pick some pixels in
        # the image for coloring them black
        # Pick a random number between 300 and 10000
        number_of_pixels = random.randint(300 , 10000)
        for i in range(number_of_pixels):
            
            # Pick a random y coordinate
            y_coord=random.randint(0, row - 1)
            
            # Pick a random x coordinate
            x_coord=random.randint(0, col - 1)
            
            # Color that pixel to black
            img[y_coord][x_coord] = 0
            
        return img
    
    def backImage(self):
        if len(self.raw_images) > 0 and self.img_index > 0:
            self.img_index -= 1
        self.showImage(self.raw_images[self.img_index],self.class_ids[self.img_index])

    def forwardImage(self):
        if (len(self.raw_images) > 0) and (self.img_index < len(self.raw_images)-1):
            self.img_index += 1
        self.showImage(self.raw_images[self.img_index],self.class_ids[self.img_index])
    
app = contourApp()
app.mainloop()
