import cv2
from skimage.measure import profile_line # conda install scikit-image, conda install chardet
import tkinter as tk
from tkinter import ttk
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from PIL import Image, ImageTk
import tensorflow as tf
from scipy.stats import norm
from scipy.signal import find_peaks,peak_widths
import os
import random

from matplotlib import cm
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk\

#image_path = 'D:/Users/James/Desktop/Catherine Science 2023/Python_Apps/GalaxyClassifier/Main/images_gz2'
train_path = 'C:/GZ2/images_cntr'
image_path = 'C:/GZ2/images_gz2'
table_path = 'D:/Users/James/Desktop/Catherine Science 2023/Python_Apps/GalaxyClassifier/Main/gz2_imgtable.csv'
prep_image_path = "C:/GZ2/images_gz2_prep"

class MainApp(tk.Tk):
    def __init__(self,master=None):
        super().__init__(master)

        self.wm_title('Galaxy Classifier')
        self.wm_geometry('1700x700')

        # Variables
        self.model = None
        self.shuffle = True

        # Menu
        menubar = tk.Menu(master=self)
        
        fileMenu = tk.Menu(menubar,tearoff=0)
        menubar.add_cascade(label="File", menu=fileMenu)
        fileMenu.add_command(label='Load model', command=self.loadModel)
        fileMenu.add_command(label='Save model', command=self.saveModel)
        fileMenu.add_separator()
        fileMenu.add_command(label='Exit', command=self.destroy)

        prepMenu = tk.Menu(menubar,tearoff=0)
        menubar.add_cascade(label="Prepocessing", menu=prepMenu)
        prepMenu.add_command(label='Preprocess all images...',command=self.prepAppExec)
        self.config(menu=menubar)

        # gui objects
        mainFrame=tk.Frame(master=master)
        mainFrame.pack(fill=tk.BOTH,side=tk.TOP,expand=True)

        displayFrame=tk.Frame(master=mainFrame)
        displayFrame.pack(fill=tk.BOTH,side=tk.TOP,expand=True)

        prepFrame=tk.LabelFrame(master=mainFrame,text='Preprocessing')
        prepFrame.pack(fill=tk.BOTH,side=tk.TOP,expand=False)

        mlcontrolFrame=tk.LabelFrame(master=mainFrame,text='Machine Learning Controls')
        mlcontrolFrame.pack(fill=tk.BOTH,side=tk.TOP,expand=False)

        controlFrame=tk.LabelFrame(master=mainFrame,text='Image Controls')
        controlFrame.pack(fill=tk.BOTH,side=tk.TOP,expand=False)

        # displayFrame objects.
        self.galaxyCanvas = tk.Canvas(master=displayFrame,width=424,height=424,bg="black")
        self.galaxyCanvas.pack(side=tk.LEFT,expand=False)

        surfFigure = plt.Figure()
        self.surfPlot = surfFigure.add_subplot(projection='3d')
        self.surfaceCanvas = FigureCanvasTkAgg(surfFigure,displayFrame)
        self.surfaceCanvas.get_tk_widget().pack(fill=tk.BOTH,side=tk.LEFT,expand=False)

        profFigure = plt.Figure()
        self.profPlot = profFigure.add_subplot(2,1,1)
        self.scanPlot = profFigure.add_subplot(2,1,2)
        self.profCanvas = FigureCanvasTkAgg(profFigure,displayFrame)
        self.profCanvas.get_tk_widget().pack(fill=tk.BOTH,side=tk.LEFT,expand=False)

        # Preprocessing Frame objects.
        cropLabel = tk.Label(prepFrame,justify=tk.LEFT,text='Resize XY To:')
        cropLabel.pack(fill=tk.BOTH,side=tk.LEFT,expand=False)
        self.cropText = tk.Entry(prepFrame)
        self.cropText.pack(fill=tk.BOTH,side=tk.LEFT,expand=False)
        self.cropText.insert(0,'200')

        # controlFrame objects.
        loadButton=tk.Button(controlFrame,text='Load Images',command=self.loadImages)
        loadButton.pack(fill=tk.BOTH,side=tk.LEFT,expand=True)

        self.backButton = tk.Button(controlFrame,text='Back',command=self.backImage)
        self.backButton.pack(fill=tk.BOTH,side=tk.LEFT,expand=True)
        self.backButton['state'] = 'disabled'

        self.forwardButton = tk.Button(controlFrame,text='Forward',command=self.forwardImage)
        self.forwardButton.pack(fill=tk.BOTH,side=tk.LEFT,expand=True)
        self.forwardButton['state'] = 'disabled'

        # mlcontrolFrame objects.
        ntrainLabel = tk.Label(mlcontrolFrame,text='Training Images:')
        ntrainLabel.pack(fill=tk.BOTH,side=tk.LEFT,expand=True)
        self.ntrainText = tk.Entry(mlcontrolFrame)
        self.ntrainText.pack(fill=tk.BOTH,side=tk.LEFT,expand=True)
        self.ntrainText.insert(0,'6752')

        ntestLabel = tk.Label(mlcontrolFrame,text='Testing Images:')
        ntestLabel.pack(fill=tk.BOTH,side=tk.LEFT,expand=True)
        self.ntestText = tk.Entry(mlcontrolFrame)
        self.ntestText.pack(fill=tk.BOTH,side=tk.LEFT,expand=True)
        self.ntestText.insert(0,'1688')

        epochsLabel = tk.Label(mlcontrolFrame,text='Epochs:')
        epochsLabel.pack(fill=tk.BOTH,side=tk.LEFT,expand=True)
        self.epochsText = tk.Entry(mlcontrolFrame)
        self.epochsText.pack(fill=tk.BOTH,side=tk.LEFT,expand=True)
        self.epochsText.insert(0,'30')

        self.trainButton = tk.Button(mlcontrolFrame,text='Train',command=self.train)
        self.trainButton.pack(fill=tk.BOTH,side=tk.LEFT,expand=True)
        self.trainButton['state'] = 'disabled'
 
        self.testButton = tk.Button(mlcontrolFrame,text='Test',command=self.test)
        self.testButton.pack(fill=tk.BOTH,side=tk.LEFT,expand=True)
        self.testButton['state'] = 'disabled'

    def loadImages(self):
        # Object variables
        self.img_data = []
        self.img_class_simp = []
        self.img_index = 0

        self.n_train = int(self.ntrainText.get())
        self.n_test = int(self.ntestText.get())
        self.n_data = self.n_train + self.n_test

        # Load Pre-processed files.
        print("Loading preprocessed images...")
        file_list = os.listdir(train_path)
        random.shuffle(file_list)
        print("Loaded %i file names"%len(file_list))

        # Load desired entries in Galaxy Classification Table.
        #print("Loading Galaxy Zoo 2 Table Samples: %s..."%(table_path))
        #gz2_table = pd.read_csv(table_path,encoding='utf-8',nrows=self.n_data)
        #print("...Loading Complete.")
        #self.asset_ids_imshow = gz2_table.iloc[:,0].values
        #self.class_ids_imshow = gz2_table.iloc[:,2].values
        
        # Load Image data.
        print('Loading image data-> %s'%(train_path))
        self.class_ids = []
        for img_file in file_list[:self.n_data]:
            f = train_path + '/' + img_file
            try:
                # Load the image.
                img_gray = cv2.imread(f,cv2.IMREAD_GRAYSCALE) # Read and convert to grayscale
                self.img_data.append(img_gray)
                # Load the class.
                self.class_ids.append(img_file.split("_")[1].split(".")[0])
                #print("File:%s class:%s"%(img_file,self.class_ids[-1]))
            except:
                print("Error Loading-> %s"%(f))
        self.img_train_data = np.array(self.img_data[:self.n_train])
        #print(self.img_train_data.shape)
        self.img_test_data = np.array(self.img_data[self.n_train:])
        #print(self.img_test_data.shape)
        print('...Complete. Loaded %i images.'%(len(self.img_data)))

        # Simplify the classifier to E or S.
        #self.class_ids = gz2_table.iloc[0:self.n_data,2].values
        self.test_truths = self.class_ids[self.n_train:]
        for c in self.class_ids:
            if 'Se' in c:
                self.img_class_simp.append(2)
            elif 'S' in c:
                self.img_class_simp.append(1)
            elif 'E' in c:
                self.img_class_simp.append(0)
        n_ellipt = self.img_class_simp.count(0)
        n_spiral = self.img_class_simp.count(1)
        n_edgespiral = self.img_class_simp.count(2)
        print('Ellipticals:%d Spirals:%d Edge-Spirals:%d'%(n_ellipt,n_spiral,n_edgespiral))
        self.img_class_train = self.img_class_simp[:self.n_train]
        print("Training labels = %d"%len(self.img_class_train))
        self.img_class_test = self.img_class_simp[self.n_train:]
        print("Testing labels = %d"%len(self.img_class_test))

        # Activate Buttons
        self.forwardButton['state'] = 'normal'
        self.backButton['state'] = 'normal'
        self.trainButton['state'] = 'normal'
        
        self.showImage()

    def showImage(self):
        if len(self.img_test_data) > 0:
            # Display the image.
            #i_path = image_path+"/"+str(self.asset_ids_imshow[self.img_index])+".jpg"
            #cv_img = cv2.imread(i_path,cv2.IMREAD_GRAYSCALE) # Read and convert to grayscale
            #height,width = cv_img.shape
            #self.tk_img = ImageTk.PhotoImage(image=Image.fromarray(cv_img)) # MUST USE "self"
            #self.galaxyCanvas.create_image(0,0,anchor=tk.NW,image=self.tk_img)

            height,width = self.img_test_data[self.img_index].shape
            self.tk_img = ImageTk.PhotoImage(image=Image.fromarray(self.img_test_data[self.img_index])) # MUST USE "self"
            self.galaxyCanvas.create_image(0,0,anchor=tk.NW,image=self.tk_img)      

            # Show cross-sectional pre-processing.
            #self.cross_sect_prep(cv_img)
            
            # Draw image cropping square.
            self.sq_xy = int(self.cropText.get())
            xy1 = width/2-self.sq_xy/2
            xy2 = width/2+self.sq_xy/2
            self.galaxyCanvas.create_rectangle(xy1,xy1,xy2,xy2,outline="light green",width=3)

            # Draw crosshair.
            line_len = 10
            self.galaxyCanvas.create_line(int(width/2-line_len/2+1),int(height/2),int(width/2+line_len/2),int(height/2),fill='light green',width=3) # Horizontal Crosshair.
            self.galaxyCanvas.create_line(int(width/2),int(height/2-line_len/2+1),int(width/2),int(height/2+line_len/2),fill='light green',width=3) # Vertical Crosshair.

            # Run prediction
            if self.model != None:
                self.sq_xy = int(self.cropText.get())
                X_pred_np = np.array(self.img_test_data[self.img_index])
                i_xy = X_pred_np.shape[1]
                X_pred_npc,y = self.preprocess(X_pred_np[int(i_xy/2-self.sq_xy/2):int(i_xy/2+self.sq_xy/2),int(i_xy/2-self.sq_xy/2):int(i_xy/2+self.sq_xy/2)],None)
                #X_pred_npc = tf.expand_dims(X_pred_np[int(i_xy/2-self.sq_xy/2):int(i_xy/2+self.sq_xy/2),int(i_xy/2-self.sq_xy/2):int(i_xy/2+self.sq_xy/2)],axis=-1)
                X_pred_npc = tf.expand_dims(X_pred_npc, axis=-1)
                X_pred_npc = tf.expand_dims(X_pred_npc, axis=0)
                pred = self.model.predict(X_pred_npc)
                # Show the class label.
                if np.argmax(pred[0]) == 0:
                    pred_text = 'Elliptical:%.2f'%(pred[0,0])
                elif np.argmax(pred[0]) == 1:
                    pred_text = 'Spiral/Disk: %.3f'%(pred[0,1])
                elif np.argmax(pred[0]) == 2:
                    pred_text = 'Edge-on Spiral/Disk: %.3f'%(pred[0,1])
            else:
                pred_text = 'None'
            #self.galaxyCanvas.create_text(5,5,anchor=tk.NW, text='%s'%(i_path),fill='red',font='Helvetica 12 bold',justify='left')
            self.galaxyCanvas.create_text(5,25,anchor=tk.NW, text='Truth: %s'%(self.test_truths[self.img_index]),fill='orange',font='Helvetica 12 bold',justify='left')
            #self.galaxyCanvas.create_text(5,25,anchor=tk.NW, text='Truth: %s'%(self.img_class_test[self.img_index]),fill='orange',font='Helvetica 12 bold',justify='left')
            self.galaxyCanvas.create_text(5,45,anchor=tk.NW, text='Predict: %s'%(pred_text),fill='turquoise',font='Helvetica 12 bold',justify='left')
            #self.galaxyCanvas.create_text(5,65,anchor=tk.NW, text='Preprocessing: %i'%(self.scan_angle[self.max_ind]),fill='purple',font='Helvetica 12 bold',justify='left')
             
            # Plot image on surface plot.
            self.surfPlot.cla()
            X = np.arange(width,0,-1)
            Y = np.arange(0,height,1)
            X,Y = np.meshgrid(X,Y)
            Z = self.img_test_data[self.img_index]
            surf = self.surfPlot.plot_surface(X,Y,Z)
            self.surfPlot.set_xlabel('X')
            self.surfPlot.set_ylabel('Y')
            self.surfPlot.set_zlabel('Intensity')
            self.surfaceCanvas.draw()
        else:
            print('No images to show.')

    def train(self):
        if len(self.img_class_train) and len(self.img_train_data):
            # Prepare the training data.
            print("Loading label data to Tensorflow.")
            train_labels = np.array(self.img_class_train)
            print("train_labels.shape:")
            print(train_labels.shape)
            self.sq_xy = int(self.cropText.get())
            X_train_np = np.array(self.img_train_data)
            print("X_train_np.shape")
            print(X_train_np.shape)
            i_xy = X_train_np.shape[1]
            X_train_npc = X_train_np[:,int(i_xy/2-self.sq_xy/2):int(i_xy/2+self.sq_xy/2),int(i_xy/2-self.sq_xy/2):int(i_xy/2+self.sq_xy/2)]
            print("X_train_npc.shape:")
            print(X_train_npc.shape)
            print("Complete.")

            # Define the training pipelines.
            tf.random.set_seed(1)
            batch_size = 1
            train_dataset = tf.data.Dataset.from_tensor_slices((X_train_npc,train_labels))
            train_dataset = train_dataset.map(self.preprocess,num_parallel_calls=tf.data.AUTOTUNE)
            train_dataset = train_dataset.cache()
            train_dataset = train_dataset.shuffle(len(train_dataset))
            train_dataset = train_dataset.batch(batch_size)
            train_dataset = train_dataset.prefetch(tf.data.AUTOTUNE)
            print(train_dataset)

            # Prepare the test data.
            test_labels = np.array(self.img_class_test)
            print("test_labels.shape:")
            print(test_labels.shape)
            X_test_np = np.array(self.img_test_data)
            i_xy = X_test_np.shape[1]
            X_test_npc = X_test_np[:,int(i_xy/2-self.sq_xy/2):int(i_xy/2+self.sq_xy/2),int(i_xy/2-self.sq_xy/2):int(i_xy/2+self.sq_xy/2)]
            print("X_test_npc.shape")
            print(X_test_npc.shape)
            
            # Define the test pipelines.
            test_dataset = tf.data.Dataset.from_tensor_slices((X_test_npc,test_labels))
            test_dataset = test_dataset.map(self.preprocess, num_parallel_calls=tf.data.AUTOTUNE)
            test_dataset = test_dataset.batch(batch_size)
            test_dataset = test_dataset.cache()
            test_dataset = test_dataset.prefetch(tf.data.AUTOTUNE)
            
            # Construct the CNN.
            if self.model == None:
                print("Generating the CNN Model.")
                self.model = tf.keras.Sequential()
                self.model.add(tf.keras.layers.InputLayer(input_shape=(self.sq_xy,self.sq_xy)))
                self.model.add(tf.keras.layers.Reshape((self.sq_xy,self.sq_xy,1)))
                
                self.model.add(tf.keras.layers.Conv2D(128, (3, 3), activation = 'relu'))
                #self.model.add(tf.keras.layers.Conv2D(64, (3, 3), activation = 'relu'))
                self.model.add(tf.keras.layers.MaxPooling2D((2, 2)))
                #self.model.add(tf.keras.layers.Dropout(rate=0.5))
                
                self.model.add(tf.keras.layers.Conv2D(128, (3, 3), activation = 'relu'))
                #self.model.add(tf.keras.layers.Conv2D(64, (3, 3), activation = 'relu'))
                self.model.add(tf.keras.layers.MaxPooling2D((2, 2)))
                #self.model.add(tf.keras.layers.Dropout(rate=0.5))
                
                self.model.add(tf.keras.layers.Conv2D(128, (3, 3), activation = 'relu'))
                self.model.add(tf.keras.layers.MaxPooling2D((2, 2)))
                
                self.model.add(tf.keras.layers.Flatten())
                
                self.model.add(tf.keras.layers.Dense(256, activation = 'relu'))
                self.model.add(tf.keras.layers.Dense(256, activation = 'relu'))
                self.model.add(tf.keras.layers.Dense(128, activation = 'relu'))
                self.model.add(tf.keras.layers.Dense(3, activation = 'softmax'))
                
                self.model.compile(optimizer='adam',\
                                  loss='sparse_categorical_crossentropy',\
                                  metrics=['accuracy'])
            else:
                self.model.summary()
                #es = tf.keras.callbacks.EarlyStopping(patience=10)
                # Train the CNN.
                print("Training the CNN.")
                epochs = int(self.epochsText.get())
                history = self.model.fit(train_dataset,\
                                         #batch_size=1,\
                                         epochs=epochs,\
                                         validation_data=test_dataset)#,\
                                         #callbacks=[es])
                print("Training complete.")
                hist=history.history
                fig = plt.figure(figsize=(13,5))
                ax = fig.add_subplot(1,2,1)
                ax.plot(hist['accuracy'],'b.-', lw=2)
                ax.set_title('Training Accuracy')
                ax.set_xlabel('Epoch')
                ax.grid()
                ax = fig.add_subplot(1,2,2)
                ax.plot(hist['val_accuracy'],'b.-', lw=2)
                ax.set_title('Validation Accuracy')
                ax.set_xlabel('Epoch')
                ax.grid()
                plt.show()

            self.testButton['state'] = 'normal'
            self.showImage()
        else:
            print("No X or Y to train.")

    def test(self):
        if self.model:
            print("Running test images.")
            self.sq_xy = int(self.cropText.get())
            X_test_np = np.array(self.img_test_data)
            i_xy = X_test_np.shape[1]
            X_test_npc = tf.expand_dims(X_test_np[:,int(i_xy/2-self.sq_xy/2):int(i_xy/2+self.sq_xy/2),int(i_xy/2-self.sq_xy/2):int(i_xy/2+self.sq_xy/2)],axis=-1)
            print(X_test_npc.shape)
            test_labels = np.array(self.img_class_test)
            print(test_labels.shape)
            results = self.model.evaluate(X_test_npc, test_labels, verbose=2)
            print("Complete.")


    def preprocess(self,image,label):
        image = tf.cast(image, tf.float32) / 255.0

        return (image,label)

    def cross_sect_prep(self, image):
        print("In cross_sect_prep()")

        # Normalize image.
        np_image = np.array(image)/255.0

        # Image properties.
        dims = np_image.shape
        rows = dims[0]
        cols = dims[1]
        print("Image rows=%d cols=%d"%(rows,cols))

        self.scan_angle = []
        self.scan_peaks = []
        self.scan_widths = []
        for a in range(-90,90):
            xi,yi,xf,yf = self.angular_scan(a,rows,cols)
            
            # Extract slice profile.
            slice_profile = profile_line(np_image,(yi,xi),(yf,xf))

            # Run statistics on the profile.
            peaks, _ = find_peaks(slice_profile)
            results_half = peak_widths(slice_profile, peaks, rel_height=0.5)
            self.max_ind = np.argmax(results_half[0])
            #print("Peaks...")
            #print(peaks)
            #print("Widths...")
            #print(results_half[0])
            self.scan_angle.append(a)
            self.scan_peaks.append(slice_profile[peaks[self.max_ind]])
            self.scan_widths.append(results_half[0][self.max_ind])
            #print("angle=%i argmax=%i width=%f peak=%f"%(a,self.max_ind,results_half[0][self.max_ind],slice_profile[peaks[self.max_ind]]))
        
        #mu, std = norm.fit(slice_profile)
        #print("mu=%f std=%f"%(mu,std))
        #x = np.linspace(-1, 1, len(slice_profile))
        #gauss = norm.pdf(x, mu, std)
        #print(gauss)

        # Plot scan characteristics.
        self.profPlot.cla()
        self.profPlot.plot(slice_profile,color="purple")
        #self.profPlot.plot(gauss)
        self.profPlot.grid()

        self.scanPlot.cla()
        self.scanPlot.plot(self.scan_angle,self.scan_widths)
        self.scanPlot.grid()
        
        self.profCanvas.draw()

        # Display cross-section cut.
        max_width_ind = np.argmax(self.scan_widths)
        xi,yi,xf,yf = self.angular_scan(self.scan_angle[max_width_ind],rows,cols)
        self.galaxyCanvas.create_line(xi,yi,xf,yf,fill='purple',width=3)
        
        # Rotate galaxy along lowest statitical variation axis?
        center = (cols/2,rows/2)
        rotate_matrix = cv2.getRotationMatrix2D(center=center, angle=self.scan_angle[max_width_ind], scale=1)
        #print(rotate_matrix)
        rotated_image = cv2.warpAffine(src=image, M=rotate_matrix, dsize=(cols, rows))

        # Calculate filtering ellipse?
        
        print("cross_sect_prep() ends.")
        
    def angular_scan(self,angle,rows,cols):
        # Cross-section line as function of scan angle.
        #angle = 60                  # Angular scan from 90 to -90.
        m = np.tan(angle*np.pi/180) # Slope from angle.
        #print("Scan line: angle=%f slope=%f"%(angle,m))
        if angle == 0:
            xi = cols
            xf = 0
            yf = yi = rows/2
        elif angle > 45:
            y0 = rows/2         # max y.
            x0 = y0/m           # corresponding x.
            xi = int(cols/2+x0) # image offset x-initial.
            xf = int(cols/2-x0) # image offset x-final.
            yi = 0
            yf = cols
        elif angle < -45:
            y0 = rows/2
            x0 = y0/m
            xi = int(cols/2-x0)
            xf = int(cols/2+x0)
            yi = cols
            yf = 0 
        else: #  -45 <= angle <= 45
            xi = cols
            xf = 0
            b0 = m*(-rows/2)
            yi = int(rows/2 + b0)
            yf = int(rows/2 - b0)

        return xi,yi,xf,yf

    def clamp(self, n, smallest, largest): return max(smallest, min(n, largest))

    def gaussian(x, alpha, r):
          return 1./(math.sqrt(alpha**np.pi))*np.exp(-alpha*np.power((x - r), 2.))

    def backImage(self):
        if len(self.img_test_data) > 0 and self.img_index > 0:
            self.img_index -= 1
        self.showImage()

    def forwardImage(self):
        if (len(self.img_test_data) > 0) and (self.img_index < len(self.img_test_data)-1):
            self.img_index += 1
        self.showImage()

    def saveModel(self):
        if self.model:
            fp = tk.filedialog.asksaveasfilename(initialdir='/',title='Save Model as',filetypes=(('h5 files','*.h5'),('all files','*.*')))
            self.model.save(fp,overwrite=True,include_optimizer=True,save_format='h5')

    def loadModel(self):
        fp = tk.filedialog.askopenfilename(initialdir='/',title='Open Saved Model',filetypes=(('h5 files','*.h5'),('all files','*.*')))
        if fp:
            self.model = tf.keras.models.load_model(fp)
            self.testButton['state'] = 'normal'
        else:
            print("Nothing to load")

    def prepAppExec(self):
        p_app = prepApp()

class prepApp(tk.Tk):
    def __init__(self,master=None):
        super().__init__(master)

        # UI Setup
        self.wm_title('Mass Image Preprocessing')
        self.wm_geometry('848x500')

        mainFrame=tk.Frame(master=self)
        mainFrame.pack(fill=tk.BOTH,side=tk.TOP,expand=True)

        displayFrame=tk.Frame(master=mainFrame)
        displayFrame.pack(fill=tk.BOTH,side=tk.TOP,expand=True)

        controlFrame=tk.Frame(master=mainFrame)
        controlFrame.pack(fill=tk.BOTH,side=tk.TOP,expand=False)

        self.origCanvas = tk.Canvas(master=displayFrame,width=424,height=424,bg="black")
        self.origCanvas.pack(side=tk.LEFT,expand=False)
        self.procCanvas = tk.Canvas(master=displayFrame,width=424,height=424,bg="black")
        self.procCanvas.pack(side=tk.LEFT,expand=False)

        self.progBar = ttk.Progressbar(controlFrame,orient='horizontal',mode='indeterminate')
        self.progBar.pack(fill=tk.BOTH,side=tk.LEFT,expand=True)
        processButton=tk.Button(controlFrame,text='Preprocess Entire Library',command=self.loadOrigImages)
        processButton.pack(fill=tk.BOTH,side=tk.LEFT,expand=False)
        cancelButton=tk.Button(controlFrame,text='Cancel',command=self.cancelLoad)
        cancelButton.pack(fill=tk.BOTH,side=tk.LEFT,expand=False)

    def cancelLoad(self):
        self.cancel = True

    def loadOrigImages(self):
        self.cancel = False
        # Load desired entries in Galaxy Classification Table.
        print("Loading Galaxy Zoo 2 Table Samples: %s..."%(table_path))
        gz2_table = pd.read_csv(table_path,encoding='utf-8')
        print("...Loading Complete.")
        asset_ids = gz2_table.iloc[:,0].values
        class_ids = gz2_table.iloc[:,2].values
        # Load original image data.
        print('Loading image data from: %s'%(image_path))
        count=0
        total = len(asset_ids)
        for img_file,class_id in zip(asset_ids,class_ids):
            image_fp = image_path+'/'+str(img_file)+'.jpg'
            try:
                fp = "%s/%s_%s.jpg"%(prep_image_path,img_file,class_id)
                # Read and convert to grayscale.
                img_gray = cv2.imread(image_fp,cv2.IMREAD_GRAYSCALE)
                # Preprocess the image.
                img_prep = self.rotate_image(img_gray)
                # Write preprocessed image to file.
                cv2.imwrite(fp,img_prep)
                print("Image file %i/%i out-> %s"%(count,total,fp))
            except Exception as e:
                print(e)
                #print("Image file not found-> %s"%(image_fp))
                continue
            count += 1
            if self.cancel == True:
                break
        print('...Image preprocessing complete.')

    def rotate_image(self, image):
        # Normalize image.
        np_image = np.array(image) #/255.0

        # Image properties.
        dims = np_image.shape
        rows = dims[0]
        cols = dims[1]
        #print("Image rows=%d cols=%d"%(rows,cols))

        self.scan_angle = []
        self.scan_peaks = []
        self.scan_widths = []
        for a in range(-90,90):
            # Extract slice profile.
            xi,yi,xf,yf = self.angular_scan(a,rows,cols)
            slice_profile = profile_line(np_image,(yi,xi),(yf,xf))
            # Find profile characteristics.
            peaks, _ = find_peaks(slice_profile)
            results_half = peak_widths(slice_profile, peaks, rel_height=0.5)
            self.max_ind = np.argmax(results_half[0])
            #print("Peaks...")
            #print(peaks)
            #print("Widths...")
            #print(results_half[0])
            self.scan_angle.append(a)
            self.scan_peaks.append(slice_profile[peaks[self.max_ind]])
            self.scan_widths.append(results_half[0][self.max_ind])
            #print("angle=%i argmax=%i width=%f peak=%f"%(a,self.max_ind,results_half[0][self.max_ind],slice_profile[peaks[self.max_ind]]))
        
        # Rotate galaxy image.
        max_width_ind = np.argmax(self.scan_widths)
        print("Rotate-> argmax=%i angle-at-max-width=%i max_width=%f "%(max_width_ind,self.scan_angle[max_width_ind],self.scan_widths[max_width_ind]))
        center = (cols/2,rows/2)
        rotate_matrix = cv2.getRotationMatrix2D(center=center, angle=-self.scan_angle[max_width_ind], scale=1)
        rotated_image = cv2.warpAffine(src=image, M=rotate_matrix, dsize=(cols, rows))

        # Calculate filtering ellipse?

        return rotated_image

    def angular_scan(self,angle,rows,cols):
        # Cross-section line as function of scan angle.
        #angle = 60                  # Angular scan from 90 to -90.
        m = np.tan(angle*np.pi/180) # Slope from angle.
        #print("Scan line: angle=%f slope=%f"%(angle,m))
        if angle == 0:
            xi = cols
            xf = 0
            yf = yi = rows/2
        elif angle > 45:
            y0 = rows/2         # max y.
            x0 = y0/m           # corresponding x.
            xi = int(cols/2+x0) # image offset x-initial.
            xf = int(cols/2-x0) # image offset x-final.
            yi = 0
            yf = cols
        elif angle < -45:
            y0 = rows/2
            x0 = y0/m
            xi = int(cols/2-x0)
            xf = int(cols/2+x0)
            yi = cols
            yf = 0 
        else: #  -45 <= angle <= 45
            xi = cols
            xf = 0
            b0 = m*(-rows/2)
            yi = int(rows/2 + b0)
            yf = int(rows/2 - b0)

        return xi,yi,xf,yf
        

app = MainApp()
app.mainloop()
